package Apache::UI;

# $Id: UI.pm,v 1.8 2001/09/22 11:59:50 xi Exp $

###############################################################################
#    Usenet Indexer - A nice and fast usenet indexer
#    Copyright (C) 2000  Christian Laursen <xi@borderworlds.dk>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###############################################################################


use strict;
use Apache::Constants qw(:common);
use CGI;
use URI::Escape;
use HTML::Entities;
use Deja;
use Data::Dumper;

sub handler {
    my $r = shift;

    chdir($r->document_root);
    readconfig;

    my $q = new CGI(scalar($r->args));

    $r->content_type("text/html");

    my $body;
    my %vars = (hostname => $r->hostname);

    if ($r->uri eq '/') {
	$vars{searchbox} = readfile("searchbox");
	$body = readfile("index.html");

	my $limits = '';
	foreach my $limit (sort {$a <=> $b} keys %limits) {
	    $limits .= qq!<option value="$limit">$limits{$limit}</option>!;
	}
	my $groups = '';
	foreach my $group (@allgroups) {
	    $groups .= qq!<option value="$group">$group</option>!;
	}
	substitute(\$vars{searchbox},{limits=>$limits,groups=>$groups});	

	substitute(\$body,\%vars);	
    } elsif ($r->uri eq '/default.css') {
	$r->content_type("text/css");
	$body = readfile("default.css");
    } elsif ($r->uri =~ m!^/search(?:\.html)?$!) {
	my $search = $q->param('search');
	my $maxage = $q->param('maxage');
	my @groups = $q->param('groups');
	my $offset = $q->param('offset');

	$search = '' unless defined $search;
	$maxage = 0 unless defined $maxage;
	@groups = () unless @groups;
	$offset = 0 unless defined $offset;

	@groups = () if @groups == @allgroups;

	$vars{search} = $search;

	my $resfile = search(\$search, \@groups, $maxage);
	$vars{results} = build_results($resfile, $search, $maxage, $offset, \@groups);

	$body = readfile("search.html");
	substitute(\$body,\%vars);
    } elsif ($r->uri =~ m!^/show(?:\.html)?$!) {
	my $id = $q->param('id');

	$vars{article} = build_article($id);

	$body = readfile("show.html");
	substitute(\$body,\%vars);
    } elsif ($r->uri =~ m!^/showthread(?:\.html)?$!) {
	my $id = $q->param('id');

	$vars{thread} = build_thread($id);

	$body = readfile("showthread.html");
	substitute(\$body,\%vars);
    } elsif ($r->uri =~ m!^/binfo(?:\.html)?$!) {
    } elsif ($r->uri =~ m!^/bsearch(?:\.html)?$!) {
	my $search = $q->param('search');
	my $maxage = $q->param('maxage');
	my @groups = $q->param('groups');
	my $offset = $q->param('offset');
	my $generation = $q->param('generation');

	$search = '' unless defined $search;
	$maxage = 0 unless defined $maxage;
	@groups = () unless @groups;
	$offset = 0 unless defined $offset;
	$generation = 1 unless defined $generation && $generation =~ /^\d+$/;

	$vars{search} = $search;

	my $resfile = search(\$search, \@groups, $maxage);
	$body = build_bresults($resfile, $offset, $generation, $search);
	$r->content_type('text/plain');
    } elsif ($r->uri =~ m!^/bshow(?:\.html)?$!) {
	my $id = $q->param('id');

	$body = build_barticle($id);
	$r->content_type('text/plain');
    } elsif ($r->uri =~ m!^/bthread(?:\.html)?$!) {
	my $id = $q->param('id');

	$body = build_bthread($id);
	$r->content_type('text/plain');
    } else {
	$r->status(404);
	$body = readfile("404.html");
#	$body = $r->uri;
    }

    $r->send_http_header;
    return OK if $r->header_only;
    $r->print($body);

    OK;
}

sub substitute {
    my ($body,$vars) = @_;

    while (my ($var,$value) = each %$vars) {
	$$body =~ s/<%\s*$var\s*%>/$value/g;
    }
}

sub readfile {
    my ($filename) = @_;

    open (FILE, $filename) or die "Couldn't read file $filename: $!";
    my $res = join ('', <FILE>);
    close FILE;
    return $res;
}

############################################################################
#                                                                          #
# Code below this point generates the dynamic content on the various pages #
#                                                                          #
############################################################################

sub build_results {
    my ($resfile,$search,$maxage,$offset,$groups) = @_;

    my $results = '';

    my $groupstring = '';
    foreach my $group (@$groups) {
	$groupstring .= "&groups=$group";
    }

    my $buffer;
    my $searchtime;
    local *FILE;
    if (open (FILE, "<$resfile") && (stat($resfile))[7]) {
	read (FILE, $buffer, 16);
	my ($secs1,$usecs1,$secs2,$usecs2) = unpack("LLLL", $buffer);
	my $secs = $secs2 - $secs1;
	my $usecs = $usecs2 - $usecs1;
	if ($usecs < 0) {
	    $secs--;
	    $usecs += 1000000;
	}
	$searchtime = sprintf ("%d.%06d", $secs, $usecs);
	$results .= "Found ".(((stat($resfile))[7]-16) / 8)." results (showing from $offset) (Search took $searchtime seconds)\n";
	if (((stat($resfile))[7]-16) / 8 == 500) {
	    $results .= "<em>(WARNING: Your search was not specific enough - results may be unreliable)</em>\n";
	}
	$results .= "<table><tr><th>From</th><th>Subject</th><th>Groups</th><th>Score</th><th>&nbsp;</th></tr>\n";
	if ($offset =~ /^\d+$/) {
	    seek (FILE, $offset * 8 + 16, 0);
	}
	my $i = 0;
	my $show = 25;
	while (read (FILE, $buffer, 8) && $i++ < $show) {
	    my ($id,$hits) = unpack("II", $buffer);
	    my $hfile = sprintf("../data/article/%08x.headers", $id);
	    my ($from, $subject, $newsgroups);
	    open (FILE2, "<$hfile") or die "Couldn't read file: $!";
	    while (<FILE2>) {
		if (m/^From: (.*)$/) {
		    $from = $1;
		} elsif (m/^Subject: (.*)$/) {
		    $subject = $1;
		} elsif (m/^Groups: (.*)$/) {
		    $newsgroups = $1;
		}
	    }
	    close FILE2;
	    $results .= "<tr><td>$from</td><td><a href=\"show?id=$id\">$subject</a></td><td>$newsgroups</td><td>$hits</td>".
		"<td><a href=\"showthread?id=$id\">Show thread</a></td></tr>\n";
	}
	$results .= "<tr><td colspan=\"4\">\n";

	if ($offset != 0) {
	    $results .= "[<a href=\"search?search=".uri_escape($search, " +\x00-\x1f\x7f-\xff")."&offset=".($offset-$show)."&maxage=$maxage$groupstring\">Previous</a>] ";
	}
	if ($i > $show && read (FILE, $buffer, 8)) {
	    $results .= "[<a href=\"search?search=".uri_escape($search, " +\x00-\x1f\x7f-\xff")."&offset=".($offset+$show)."&maxage=$maxage$groupstring\">Next</a>]";
	}
	close FILE;

	$results .= "</td></tr>\n</table>\n";
    } else {
	$results .= "No results found.";
    }

    return $results;
}

sub build_article {
    my ($id) = @_;

    my $article = '';

    if (defined $id && $id =~ /^\d+$/) {
	$article .= "<pre>\n";
	my $prefix = sprintf("../data/article/%08x", $id);
	my $parent = undef;
	open (FILE, "<$prefix.headers");
	while (<FILE>) {
	    s/^Time: (\d+)$/'Date: '.localtime($1)/e;
	    if (m/^Parent: (\d+)$/) {
		$parent = $1;
	    } else {
		$article .= encode_entities($_);
	    }
	}
	close FILE;
	my @children = ();
	if (open (FILE, "<$prefix.children")) {
	    my $buffer;
	    while (read (FILE, $buffer, 4)) {
		push @children, unpack('I',$buffer);
	    }
	    close FILE;
	}
	$article .= "</pre>\n";
	if (defined $parent) {
	    $article .= qq!<a href="show?id=$parent">Previous in thread</a><br>\n!;
	}
	if (@children > 0) {
	    $article .= "Next in thread: ";
	    for (my $i = 0; $i < @children; ++$i) {
		$article .= "<a href=\"show?id=$children[$i]\">".($i+1)."</a> ";
	    }
	    $article .= "<br>\n";
	}
	$article .= "<a href=\"showthread?id=$id\">Show entire thread</a><br>\n";
	$article .= "<hr>\n<pre>\n";

	open (FILE, "<$prefix.body");
	while (<FILE>) {
	    my $tmp = substr($_,0,10);
	    $tmp =~ s/[^>]//g;
	    if (length($tmp) > 0) {
		chomp;
		$article .= '<span class="quoted'.length($tmp).'">'.encode_entities($_)."</span>\n";
	    } else {
		$article .= encode_entities($_);
	    }
	}
	close FILE;
	$article .= "</pre>\n";
    } else {
	$article .= "You evil bastard!\n";
    }   
}

sub build_thread {
    my ($id) = @_;

    my $thread = '';

    if ($id !~ /^\d+$/) {
	return "<html><body>You bastard!</body></html>";
    }

# Find the top article

    my $top = $id;
    while (my $parent = parent($top)) {
	$top = $parent;
    }

    show_hier(\$thread, $top);

    return $thread;
}

sub show_hier {
    my $thread = shift;
    $$thread .= "<ul>\n";
    foreach my $id (@_) {
	my ($from, $subject) = get_headers($id);
	$$thread .= "<li><a href=\"show?id=$id\">$from, $subject</a>\n";
	my @children = children($id);
	show_hier($thread,@children) if (@children > 0);
	$$thread .= "</li>\n";
    }
    $$thread .= "</ul>\n";
}

sub build_bresults {
    my ($resfile, $offset, $generation, $search) = @_;

    my $results = '';

    my $buffer;
    my $searchtime;
    if (open (FILE, "<$resfile") && (stat($resfile))[7]) {
	read (FILE, $buffer, 16);
	my ($secs1,$usecs1,$secs2,$usecs2) = unpack("LLLL", $buffer);
	my $secs = $secs2 - $secs1;
	my $usecs = $usecs2 - $usecs1;
	if ($usecs < 0) {
	    $secs--;
	    $usecs += 1000000;
	}
	$searchtime = sprintf ("%d.%06d", $secs, $usecs);

	$results .= 'Results: '.(((stat($resfile))[7]-16) / 8)."\n";
	$results .= "Searchtime: $searchtime\n";

	if ($offset =~ /^\d+$/) {
	    seek (FILE, $offset * 8 + 16, 0);
	}
	my $i = 0;
	my $show = 25;

	while (read (FILE, $buffer, 8) && $i++ < $show) {
	    my ($id,$hits) = unpack("II", $buffer);
	    my $hfile = sprintf("../data/article/%08x.headers", $id);
	    my ($from, $subject, $newsgroups);
	    open (FILE2, "<$hfile") or die "Couldn't read file: $!";
	    while (<FILE2>) {
		if (m/^From: (.*)$/) {
		    $from = $1;
		} elsif (m/^Subject: (.*)$/) {
		    $subject = $1;
		} elsif (m/^Groups: (.*)$/) {
		    $newsgroups = $1;
		}
	    }
	    close FILE2;
	    $results .= "Id: $id\n";
	    $results .= "From: $from\n";
	    $results .= "Subject: $subject\n";
	    $results .= "Score: $hits\n";
	    if ($generation >= 2) {
		$results .= "Groups: $newsgroups\n";

		my @snips = ();
		my @words = ();
		while ($search =~ m/[+\-]?([\w$natchars]+)/g) {
		    push @words, $1;
		}
		my $bfile = sprintf("../data/article/%08x.body", $id);
		open (FILE2, "<$bfile") or die "Couldn't read file: $!";
		while (<FILE2>) {
		    my $line = lc2($_);
		    foreach my $word (@words) {
			if ($line =~ m/$word/) {
			    push @snips, $_;
			}
		    }
		}
		close FILE2;

		$results .= "Snips: ".scalar(@snips)."\n";
		foreach my $snip (@snips) {
		    $results .= "Snip: $snip";
		}
	    }
	}
	close FILE;
    }

    return $results;
}

sub build_barticle {
    my ($id) = @_;

    my $article = '';

    if ($id =~ /^\d+$/) {
	my $prefix = sprintf("../data/article/%08x", $id);
	open (FILE, "<$prefix.headers") or die "Coulnd't read file: $!";
	while (<FILE>) {
	    $article .= $_;
	}
	close FILE;
	if (open (FILE, "<$prefix.children")) {
	    my $buffer;
	    my @children = ();
	    while (read(FILE, $buffer, 4)) {
		push @children, unpack('I', $buffer);
	    }
	    close FILE;
	    $article .= "Children: ".join(',',@children)."\n";
	}
	$article .= "\n";
	open (FILE, "<$prefix.body");
	while (<FILE>) {
	    $article .= $_;
	}
	close FILE;
    } else {
	$article .= 'You evil bastard!';
    }

    return $article;
}

sub build_bthread {
    my ($id) = @_;

    my $thread = '';

    if ($id !~ /^\d+$/) {
	return "You bastard!";
    }

# Find the top article

    my $top = $id;
    while (my $parent = parent($top)) {
	$top = $parent;
    }

    show_hier2(\$thread, 0, $top);

    return $thread;
}

sub show_hier2 {
    my $thread = shift;
    my $level = shift;
    foreach my $id (@_) {
	my ($from, $subject) = get_headers($id);
	$$thread .= "Id: $id\n";
	$$thread .= "From: $from\n";
	$$thread .= "Subject: $subject\n";
	$$thread .= "Level: $level\n";
	my @children = children($id);
	show_hier2($thread, $level + 1, @children) if (@children > 0);
    }
}

1;
